function [fid, refs] = sortMeas_TXfct(mrprot, mdh, rawfid, type)
% sort raw fid data from Siemens meas.out/meas.asc files
% into standard 8D space (points,lines,slices,partitions,contrasts,channels,reps)
% for segmented, create pcorfid and index to describe which
%   correction goes with which line
% for iPAT, create patref for autocalibration lines
%
% usage: [fid, refs] = sortMeas(mrprot, mdh, rawfid, {'epi','spec','radial','tse'(default)}
%
% ** NEW ** refs is now returned as a struct
%
% for TSE/GRE, refs = {coil_names, noisefid[, patref]}
% for radial, refs = {coil_names, noisefid, angles}
% for EPI, refs = {coil_names, noisefid, pcor[, patref]}
% for EPI+multi-band, refs = {coil_names, noisefid, pcor, sbfid, sbpcor[, patref, sbpatref]}
% for spectroscopy, refs = {coil_names, noisefid, reffid}
%
% dependencies: none

version = '2013.07.17';

fprintf('\nsortMeas.m version %s by eja\n------------------------------------------\n', version);

% definition of EvalInfoMask:
MDH_ACQEND            =  1;
MDH_RTFEEDBACK        =  2;
MDH_HPFEEDBACK        =  3;
%MDH_ONLINE            =  4;
%MDH_OFFLINE           =  5;
%MDH_SYNCDATA          =  6;      % readout contains synchroneous data
%MDH_LASTSCANINCONCAT  =  9;      % Flag for last scan in concatination
%MDH_RAWDATACORRECTION = 11;      % Correct the rawadata with the rawdata correction factor
%MDH_LASTSCANINMEAS    = 12;      % Flag for last scan in measurement
%MDH_SCANSCALEFACTOR   = 13;      % Flag for scan specific additional scale factor
%MDH_2NDHADAMARPULSE   = 14;      % 2nd RF exitation of HADAMAR
MDH_REFPHASESTABSCAN  = 15;      % reference phase stabilization scan
MDH_PHASESTABSCAN     = 16;      % phase stabilization scan
%MDH_D3FFT             = 17;      % execute 3D FFT
%MDH_SIGNREV           = 18;      % sign reversal
%MDH_PHASEFFT          = 19;      % execute phase fft
%MDH_SWAPPED           = 20;      % swapped phase/readout direction
%MDH_POSTSHAREDLINE    = 21;      % shared line
MDH_PHASCOR           = 22;      % phase correction data
MDH_PATREFSCAN        = 23;      % additonal scan for PAT reference line/partition
MDH_PATREFANDIMASCAN  = 24;      % additonal scan for PAT reference line/partition that is also used as image scan
%MDH_REFLECT           = 25;      % reflect line
MDH_NOISEADJSCAN      = 26;      % noise adjust scan --> used since VB13
%MDH_SHARENOW          = 27;      % all lines are acquired from the actual and previous e.g. phases
%MDH_LASTMEASUREDLINE  = 28;      % indicates that the current line is the last measured line of all succeeding e.g. phases
%MDH_FIRSTSCANINSLICE  = 29;      % indicates first scan in slice (needed for time stamps)
%MDH_LASTSCANINSLICE   = 30;      % indicates  last scan in slice (needed for time stamps)
%MDH_TREFFECTIVEBEGIN  = 31;      % indicates the begin time stamp for TReff (triggered measurement)
%MDH_TREFFECTIVEEND    = 32;      % indicates the   end time stamp for TReff (triggered measurement)
% eja consts
%MDH_SIR_FIRST_ECHO            = 37;	  % indicates first SIR slice, which may only need interpolation for ramp up
%MDH_SIR_LAST_ECHO             = 38;      % indicates last SIR slice, which may only need interpolation for ramp down
MDH_MULTI_BAND_REFSCAN        = 39;      % indicates single-band reference scan within multi-band acquisition
%MDH_MULTI_BAND_IMASCAN        = 40;      % indicates multi-band scan
% remaining siemens consts
%MDH_FIRST_SCAN_IN_BLADE       = 41;      % Marks the first line of a blade
%MDH_LAST_SCAN_IN_BLADE        = 42;      % Marks the last line of a blade
%MDH_LAST_BLADE_IN_TR          = 43;      % Set for all lines of the last BLADE in each TR interval
%MDH_RETRO_LASTPHASE           = 46;      % Marks the last phase in a heartbeat
%MDH_RETRO_ENDOFMEAS           = 47;      % Marks an ADC at the end of the measurement
%MDH_RETRO_REPEATTHISHEARTBEAT = 48;      % Repeat the current heartbeat when this bit is found
%MDH_RETRO_REPEATPREVHEARTBEAT = 49;      % Repeat the previous heartbeat when this bit is found
%MDH_RETRO_ABORTSCANNOW        = 50;      % Just abort everything
%MDH_RETRO_LASTHEARTBEAT       = 51;      % This adc is from the last heartbeat (a dummy)
%MDH_RETRO_DUMMYSCAN           = 52;      % This adc is just a dummy scan, throw it away
%MDH_RETRO_ARRDETDISABLED      = 53;      % Disable all arrhythmia detection when this bit is found

% precompute flags we will check later
isACQEND            = ( bitget(mdh.aulEvalInfoMask(:,1),MDH_ACQEND) == 1 );
isRTFEEDBACK        = ((bitget(mdh.aulEvalInfoMask(:,1),MDH_RTFEEDBACK) == 1) & (bitget(mdh.aulEvalInfoMask(:,1),MDH_PHASCOR) ~= 1) ); % not shared PHASCOR lines
isHPFEEDBACK        = ( bitget(mdh.aulEvalInfoMask(:,1),MDH_HPFEEDBACK) == 1);
isREFPHASESTABSCAN  = ( bitget(mdh.aulEvalInfoMask(:,1),MDH_REFPHASESTABSCAN) == 1 );
isPHASESTABSCAN     = ( bitget(mdh.aulEvalInfoMask(:,1),MDH_PHASESTABSCAN) == 1 );
isPHASCOR           = ( bitget(mdh.aulEvalInfoMask(:,1),MDH_PHASCOR) == 1 );
isPATREFSCAN        = ( bitget(mdh.aulEvalInfoMask(:,1),MDH_PATREFSCAN) == 1 );
isPATREFANDIMASCAN  = ( bitget(mdh.aulEvalInfoMask(:,1),MDH_PATREFANDIMASCAN) == 1 );
isNOISEADJSCAN      = ( bitget(mdh.aulEvalInfoMask(:,1),MDH_NOISEADJSCAN) == 1 );
isSINGLEBANDREF     = ( bitget(mdh.aulEvalInfoMask(:,2),MDH_MULTI_BAND_REFSCAN-32) == 1 );

% these are all the lines we want to skip
isIGNORE = isACQEND | isRTFEEDBACK | isHPFEEDBACK;

% check endian-ness of system
[~,~,endian] = computer;

isEPI = false;
isSpec = false;
isRadial = false;
if (~exist('type','var')), type = 'tse'; end
if (strcmpi(type, 'epi'))
    isEPI = true;
    fprintf('Assuming EPI sequence\n');
elseif (strcmpi(type, 'spec'))
    isSpec = true;
    fprintf('Assuming spectroscopy sequence\n');
elseif (strcmpi(type, 'radial'))
    isRadial = true;
    fprintf('Assuming radial sequence\n');
else
    fprintf('Assuming TSE/GRE sequence\n');
end

isNewEPI = false; isMultiBand = false; isSBRef = false;
if (isEPI)
    if (mrprot.alICEProgramPara(11) == 123454321)
        isNewEPI = true;
        isMultiBand = (mrprot.alICEProgramPara(12) > 1);
    elseif (mrprot.alICEProgramPara(11) == 123454322)
        isNewEPI = true;
        isMultiBand = (mrprot.alICEProgramPara(12) > 1);
        isSBRef = true;
    end
    
    iMrProtSlices = mrprot.sSliceArray.lSize;
    iActualSlices = max([iMrProtSlices  mrprot.sSpecPara.lFinalMatrixSizeSlice]);
end

isSlicesGt128 = false;
if (isNewEPI)
    % if necessary (slices > 128), calculate the chronologic <-> anatomic slice index map (previously lazily stored in TD)
    isSlicesGt128 = (iActualSlices > iMrProtSlices);
    
    if (isSlicesGt128)
        viSlcAt = -1*ones(iActualSlices,1);
        lSlicesToMeasure = iActualSlices;
        
        for lChronologicSliceNumber=0:lSlicesToMeasure-1
            if (mrprot.sSliceArray.ucMode == 2) % SEQ::DESCENDING
                lSliceIndex = lSlicesToMeasure-lChronologicSliceNumber-1;
            elseif (mrprot.sSliceArray.ucMode == 1) % SEQ::ASCENDING
                lSliceIndex = lChronologicSliceNumber;
            else % interleaved
                lSliceIndex = lChronologicSliceNumber;
                if ( (mod(lSlicesToMeasure,2) && mod(lSliceIndex,2)) || (~mod(lSlicesToMeasure,2) && ~mod(lSliceIndex,2)) )
                    %if ( (lSlicesToMeasure % 2) ? (lSliceIndex % 2) : !(lSliceIndex % 2) ) {
                    lSliceIndex = floor(lSliceIndex / 2) + floor((lSlicesToMeasure + 1) / 2);
                else
                    lSliceIndex = floor(lSliceIndex / 2);
                end
            end
            
            viSlcAt(lSliceIndex+1) = lChronologicSliceNumber;
        end
    else
        viSlcAt = mrprot.alTD;
    end
end

% NOTE/WARNING: right now this may not work for cardiac or venc

[nSamples, nTraces, nCH] = size(rawfid);

% find maxima
maxLines  = 1 + double(max(mdh.ushLine));           % 2d phase encodes
nAcqs     = 1 + double(max(mdh.ushAcquisition));    % acquisitions (averages)
nSlices   = 1 + double(max(mdh.ushSlice));          % slices
nParts    = 1 + double(max(mdh.ushPartition));      % 3d partitions
nEchos    = 1 + double(max(mdh.ushEcho));           % contrasts
nPhases   = 1 + double(max(mdh.ushPhase));          % cardiac phase (unused)
nReps     = 1 + double(max(mdh.ushRepetition));     % measurement repetitions
nSets     = 1 + double(max(mdh.ushSet));            % set (average for spectroscopy, venc for imaging, contrasts for 2xIR sequence)
nSegs     = 1 + double(max(mdh.ushSeg));            % segment (for TSE) (redundant w/ lines, unused)
nIda      = 1 + double(max(mdh.ushIda));            % free loop a
nIdb      = 1 + double(max(mdh.ushIdb));            % free loop b
nIdc      = 1 + double(max(mdh.ushIdc));            % free loop c
nIdd      = 1 + double(max(mdh.ushIdd));            % free loop d
nIde      = 1 + double(max(mdh.ushIde));            % free loop e

if (isEPI)
    if (isMultiBand)
        nIdb = 1;
        nIdc = 1;
        sliceBands = mrprot.alICEProgramPara(12);
        sliceBandInc = mrprot.alICEProgramPara(13);
        fprintf('  ==> This appears to be multi-band data (bands = %d; acquired slices = %d)\n', sliceBands, sliceBandInc);
    end
    if (isSBRef)
        nSlicesSB = nSlices;
        fprintf('  ==> Single-band reference scan present (actual slices = %d)\n', nSlicesSB);
    end
    if (isMultiBand)
        nSlices = sliceBandInc;
    end
    if (isSlicesGt128)
        fprintf('  ==> NOTE: the option to acquire slices > 128 was invoked for this dataset!\n');
    end
end

if (isRadial)
    nViews = getmrprot(mrprot.sKSpace,'lRadialViews');
    nInts = getmrprot(mrprot.sKSpace,'lRadialInterleavesPerImage');
    if (numel(mrprot.sWiPMemBlock.alFree) >= 3) % case where views are read from file
        if (mrprot.sWiPMemBlock.alFree(1+1) == 2)
            nViews = mrprot.sWiPMemBlock.alFree(2+1);
            nInts = 1;
        end
    end
    nProj = nViews * nInts;
    angles.theta = zeros(1,nProj,'single');
    angles.phi = zeros(1,nProj,'single');
    maxLines = nProj;
    nParts = 1;         % used for interleaves
    nIda = 1;           % used to wrap cline > 65535
end

% find number of active channels (VB15 fix)
if isfield(mrprot.sRXSPEC,'aFFT_SCALE')
    aFFT_SCALE = mrprot.sRXSPEC.aFFT_SCALE;
elseif isfield(mrprot, 'sCoilSelectMeas') % VD11
    if isfield(mrprot.sCoilSelectMeas, 'aRxCoilSelectData')
        aFFT_SCALE = mrprot.sCoilSelectMeas.aRxCoilSelectData(1).aFFT_SCALE;
    end
elseif isfield(mrprot, 'asCoilSelectMeas')
    if isfield(mrprot.asCoilSelectMeas(1),'aFFT_SCALE') % VB15
        aFFT_SCALE = mrprot.asCoilSelectMeas(1).aFFT_SCALE;
    end
elseif (nCH > 1)
    % ejaCplxRaw has only XProtocol, and I didn't parse the arrays
else
    error('ERROR: could not find FFT_SCALE factor in sRXSPEC or asCoilSelectMeas!');
end

if (nCH > 1) % this is a VDxx 3D array
    nChannels = nCH;
    maxChannels = 128;  % 128 is max for VDxx system
else
    nChannels = length(aFFT_SCALE);      % receiver channels/coils
    maxChannels = 32; % 32 is max for <=VBxx system
end

% see if this is a PAT scan
if ( nnz(isPATREFSCAN) || nnz(isPATREFANDIMASCAN) )
    foundpat = true;
else
    foundpat = false;
end

% error checking
if ((isSpec) && ((isEPI) || (foundpat)))
    error('ERROR: detected spectroscopy with EPI and/or phase correction?!?!');
end
if ((isRadial) && (foundpat))
    error('ERROR: radial + iPAT currently not supported!');
end

% find number of image lines actually acquired (to reduce memory allocated for ipat scans)
% also determine if we have internal or external reference
foundextpat = false;
if (foundpat)
    nImgLines = 0;
    nPatLines = 0;
    imgLineIdx = zeros(1,maxLines,'uint16');
    patLineIdx = imgLineIdx;
    patref.imgLineIdx = imgLineIdx;
    patref.patLineIdx = imgLineIdx;
    for x=1:nTraces
        if ( ~isIGNORE(x) && ~isREFPHASESTABSCAN(x) && ~isPHASESTABSCAN(x) && ~isPHASCOR(x) && ~isNOISEADJSCAN(x) )
            if ( isPATREFSCAN(x) && ~isPATREFANDIMASCAN(x) )  % external pat line
                foundextpat = true;
                if (patLineIdx(mdh.ushLine(x)+1) == 0)
                    nPatLines = nPatLines + 1;
                    patLineIdx(mdh.ushLine(x)+1) = nPatLines;
                    patref.patLineIdx(nPatLines) = mdh.ushLine(x)+1;
                end
            elseif ( ~isPATREFSCAN(x) || isPATREFANDIMASCAN(x) )  % image line or shared (internal) pat line
                if (imgLineIdx(mdh.ushLine(x)+1) == 0)
                    nImgLines = nImgLines + 1;
                    imgLineIdx(mdh.ushLine(x)+1) = nImgLines;
                    patref.imgLineIdx(nImgLines) = mdh.ushLine(x)+1;
                end
            end
        end
    end
    patref.imgLineIdx = patref.imgLineIdx(1:nImgLines);
    patref.patLineIdx = patref.patLineIdx(1:nPatLines);
else
    nImgLines = maxLines;
    imgLineIdx = 1:maxLines;
end

if (isSpec)
    if (nSlices>1), error('ERROR: found %d SLICES (I don''t know how to handle this!)',nSlices); end
    if (nParts>1), error('ERROR: found %d PARTITIONS (I don''t know how to handle this!)',nParts); end
else
    if (nPhases>1), error('ERROR: found %d PHASES (I don''t know how to handle this!)',nPhases); end
    if (nSets>1)
        % the double-echo IR sequence (eja_tfl_2contrast, newer eja_tfl) uses SETS for CONTRASTS
        % in this case, we will substitute sets for echos in this code (both shouldn't be used together...)
        if (nEchos>1)
            error('ERROR: found %d SETS AND %d ECHOES (I don''t know how to handle this!)',nSets,nEchos);
        else
            nEchos = nSets;
            mdh.ushEcho = mdh.ushSet;
            nSets = 1;
            mdh.ushSet = 0 * mdh.ushSet;
        end
    end
end

nTimePts = nReps*nIda*nIdb*nIdc*nIdd*nIde;

if (isSpec)
    arrSize = [nSamples nChannels nSets nEchos nTimePts nImgLines nPhases nSegs]; % for csi: 1d/2d/3d=Lines/Phases/Segs
else
    arrSize = [nSamples nImgLines nSlices nParts nEchos nChannels nTimePts];
end

% use mdh info to preallocate array
fprintf('Allocating memory for sorted data (complex single array):\n');
if (isSpec)
    fprintf('  %d point',arrSize(1));
    if (arrSize(1) > 1), fprintf('s'); end
    fprintf(', %d channel',arrSize(2));
    if (arrSize(2) > 1), fprintf('s'); end
    fprintf(', %d average',arrSize(3));
    if (arrSize(3) > 1), fprintf('s'); end
    fprintf(', %d echo',arrSize(4));
    if (arrSize(4) > 1), fprintf('es'); end
    fprintf(', %d repeat',arrSize(5));
    if (arrSize(5) > 1), fprintf('s'); end
    fprintf(', %d 1D PE',arrSize(6));
    fprintf(', %d 2D PE',arrSize(7));
    fprintf(', %d 3D PE',arrSize(8));
else
    fprintf('  %d point',arrSize(1));
    if (arrSize(1) > 1), fprintf('s'); end
    fprintf(', %d line',arrSize(2));
    if (arrSize(2) > 1), fprintf('s'); end
    fprintf(', %d slice',arrSize(3));
    if (arrSize(3) > 1), fprintf('s'); end
    fprintf(', %d partition',arrSize(4));
    if (arrSize(4) > 1), fprintf('s'); end
    fprintf(', %d contrast',arrSize(5));
    if (arrSize(5) > 1), fprintf('s'); end
    fprintf(', %d channel',arrSize(6));
    if (arrSize(6) > 1), fprintf('s'); end
    fprintf(', %d repeat',arrSize(7));
    if (arrSize(7) > 1), fprintf('s'); end
end
fprintf('\n');

fid = complex(zeros(arrSize,'single'));
noisefid = complex(zeros([nSamples nChannels],'single'));
if (isSpec), reffid = noisefid; end
foundNoiseFid = false;

if (isSBRef)
    sbfid = complex(zeros([nSamples nImgLines nSlicesSB nParts nEchos nChannels],'single'));
end

if (foundextpat) % for external pat reference
    arrSize = [nSamples nPatLines nSlices nParts nEchos nChannels];
    fprintf('Allocating memory for PAT reference lines (complex single array):\n');
    fprintf('  %d point',arrSize(1));
    if (arrSize(1) > 1), fprintf('s'); end
    fprintf(', %d line',arrSize(2));
    if (arrSize(2) > 1), fprintf('s'); end
    fprintf(', %d slice',arrSize(3));
    if (arrSize(3) > 1), fprintf('s'); end
    fprintf(', %d partition',arrSize(4));
    if (arrSize(4) > 1), fprintf('s'); end
    fprintf(', %d contrast',arrSize(5));
    if (arrSize(5) > 1), fprintf('s'); end
    fprintf(', %d channel',arrSize(6));
    if (arrSize(6) > 1), fprintf('s'); end
    fprintf('\n');
    patref.fid = complex(zeros(arrSize,'single'));
    
    if (isSBRef)
        sbpatref.imgLineIdx = patref.imgLineIdx;
        sbpatref.patLineIdx = patref.patLineIdx;
        sbpatref.fid = complex(zeros([nSamples nPatLines nSlicesSB nParts nEchos nChannels],'single'));
    end
end

if (isEPI)
    % find number of phase correction scans (per slice: i.e., this should only be >1 for segmented)
    if (isSBRef)
        lastwaspcor = zeros(nSlicesSB,1);
        maxpcor = zeros(nSlicesSB,1);
    else
        lastwaspcor = zeros(nSlices,1);
        maxpcor = zeros(nSlices,1);
    end
    for x=1:nTraces
        cSlc = mdh.ushSlice(x)+1;
        if (isNewEPI), cSlc = viSlcAt(cSlc) + 1; end
        if (isPHASCOR(x))
            if (lastwaspcor(cSlc) == 0)
                maxpcor(cSlc) = maxpcor(cSlc) + 1;
            end
            lastwaspcor(cSlc) = 1;
        else
            lastwaspcor = 0*lastwaspcor;
        end
    end
    
    % allocate memory for pcor, patref stuff
    if (~isSBRef)
        if (maxpcor == nTimePts*ones(nSlices,1))
            fprintf('EPI: found one phase correction block per slice ==> standard epi sequence\n');
        elseif (maxpcor > nTimePts*ones(nSlices,1))
            fprintf('EPI: found multiple phase correction blocks per slice ==> segmented epi sequence?\n');
        elseif (~isMultiBand)
            fprintf('EPI: WARNING -- some slices missing phase correction blocks!!\n');
        end
    end
    
    pcor.fid = complex(zeros([nSamples max(maxpcor) nSegs nSlices nChannels],'single'));
    pcor.idx = zeros([2 nImgLines nSlices nParts nEchos nTimePts],'uint16');
    if (foundextpat), patref.pcor_idx = zeros([2 nPatLines nSlices nParts nEchos],'uint16'); end
    
    if (isSBRef)
        sbpcor.fid = complex(zeros([nSamples max(maxpcor) nSegs nSlicesSB nChannels],'single'));
        sbpcor.idx = zeros([2 nImgLines nSlicesSB nParts nEchos],'uint16');
        if (foundextpat), sbpatref.pcor_idx = zeros([2 nPatLines nSlicesSB nParts nEchos],'uint16'); end
    end
end

test_single = fid(1); %#ok<NASGU>
single_size = whos('test_single');
single_size = single_size.bytes;
fprintf('Memory for sorted data successfully allocated (%.1f MB)\n',(numel(fid)*single_size)/1048576);
if (foundextpat) % for external pat reference
    fprintf('Memory for PAT reference lines successfully allocated (%.1f MB)\n',(numel(patref.fid)*single_size)/1048576);
end

if ((nAcqs > 1) && ~isEPI)
    fprintf('Summing %d averages\n',nAcqs);
    defaultAverage = 1;
elseif (nnz(isRTFEEDBACK))
    fprintf('** Found RTFEEDBACK lines -- assuming navigated sequence, ignoring all but last repeat\n');
    defaultAverage = -1; % ignore all but the last repeated line in navigated scans?
else
    defaultAverage = 0;
end
avgWarnStatus_spcpcor  = false;
avgWarnStatus_sbpcor   = false;
avgWarnStatus_mbpcor   = false;
avgWarnStatus_sbpatref = false;
avgWarnStatus_mbpatref = false;
avgWarnStatus_sbimg    = false;
avgWarnStatus_mbimg    = false;

% create cross-index for channel/coil/receiver info
if (1 == exist('aFFT_SCALE','var'))
    coil_idx = zeros(1,maxChannels,'uint16');
    for x=1:nChannels
        if (isfield(aFFT_SCALE,'lADCChannel')) % VD11
            coil_idx(aFFT_SCALE(x).lADCChannel) = x;
        else
            coil_idx(aFFT_SCALE(x).lRxChannel) = x;
        end
    end
    % create list of coil element names
    if (isfield(mrprot,'sCoilSelectMeas')) % VD11
        maxCoils = length(mrprot.sCoilSelectMeas.aRxCoilSelectData(1).asList);
    else
        maxCoils = length(mrprot.asCoilSelectMeas.asList);
    end
    coil_names = {};
    for x=1:maxCoils
        if (isfield(mrprot,'sCoilSelectMeas')) % VD11
            if (mrprot.sCoilSelectMeas.aRxCoilSelectData(1).asList(x).lElementSelected == 1)
                coil_names(coil_idx(mrprot.sCoilSelectMeas.aRxCoilSelectData(1).asList(x).lADCChannelConnected)) = ...
                    cellstr(mrprot.sCoilSelectMeas.aRxCoilSelectData(1).asList(x).sCoilElementID.tElement); %#ok<AGROW>(never more than 32)
            end
        else
            if (mrprot.asCoilSelectMeas.asList(x).lElementSelected == 1)
                coil_names(coil_idx(mrprot.asCoilSelectMeas.asList(x).lRxChannelConnected)) = ...
                    cellstr(mrprot.asCoilSelectMeas.asList(x).sCoilElementID.tElement); %#ok<AGROW>(never more than 32)
            end
        end
    end
end

% create index for slice positions
if (~isSpec && ~isNewEPI), sSlicePos = zeros(nSlices,3,'single'); end

% loop through data and sort
fprintf('Sorting raw data:   0%%');
if (isSBRef)
    lastwaspcor = zeros(nSlicesSB,1);
    pcorctr = zeros(nSlicesSB,1,'uint16');
else
    lastwaspcor = zeros(nSlices,1);
    pcorctr = zeros(nSlices,1,'uint16');
end
for idx=1:nTraces
    for cCH=1:nCH
        % compute indices
        cImgLine = imgLineIdx(mdh.ushLine(idx)+1);
        if (foundpat), cPatLine = patLineIdx(mdh.ushLine(idx)+1); end
        cAcq     = mdh.ushAcquisition(idx)+1;
        cSlice   = mdh.ushSlice(idx)+1;
        cPart    = mdh.ushPartition(idx)+1;
        cEcho    = mdh.ushEcho(idx)+1;
        cPhase   = mdh.ushPhase(idx)+1;
        cRep     = mdh.ushRepetition(idx)+1;
        cSet     = mdh.ushSet(idx)+1;
        cSeg     = mdh.ushSeg(idx)+1;
        cIda     = mdh.ushIda(idx)+1;
        cIdb     = mdh.ushIdb(idx)+1;
        cIdc     = mdh.ushIdc(idx)+1;
        cIdd     = mdh.ushIdd(idx)+1;
        cIde     = mdh.ushIde(idx)+1;
        
        if (isEPI && isMultiBand)
            cIdb = 1;
            cIdc = 1;
        end
        
        if (isfield(mdh, 'sCH_ulChannelId'))
            cChannel = coil_idx(mdh.sCH_ulChannelId(idx,cCH)+1);
        elseif (isfield(mdh, 'ulChannelId'))
            cChannel = coil_idx(mdh.ulChannelId(idx)+1);
        else
            cChannel = cCH; % eja for ejaCplxRaw
        end
        
        if (isNewEPI)
            cSlice = viSlcAt(cSlice) + 1;
        end
        
        if (isRadial)
            % unwrap current view index, which wraps at 65535 (ushort)
            cImgLine = mdh.ushPartition(idx)*nViews + (mdh.ushIda(idx)*65535 + mdh.ushLine(idx)) + 1;
            cPart = 1;
            cIda = 1;
            
            % angles (2*float) are stored in the space for the 4*short IceProgramPara
            theta = typecast(mdh.aushIceProgramPara(idx, 1:2), 'single');
            phi = typecast(mdh.aushIceProgramPara(idx, 3:4), 'single');
            
            if (endian == 'L') % little-endian (default)
                angles.theta(cImgLine) = theta;
                angles.phi(cImgLine) = phi;
            else
                angles.theta(cImgLine) = swapbytes(theta);
                angles.phi(cImgLine) = swapbytes(phi);
            end
        end
        
        % loop the free loops inside of the repetition loop for now
        cTimePt = (cRep-1)*(nIda+nIdb+nIdc+nIdd+nIde-5) ...
            + (cIde-1)*(nIda+nIdb+nIdc+nIdd-4) ...
            + (cIdd-1)*(nIda+nIdb+nIdc-3) ...
            + (cIdc-1)*(nIda+nIdb-2) ...
            + (cIdb-1)*(nIda-1) ...
            + cIda+cIdb+cIdc+cIdd+cIde+cRep-5;
        
        if (isIGNORE(idx))
            % skip the acqend, etc. lines
        else
            if (isNOISEADJSCAN(idx))
                % noiseadj measurements appear once per channel in VB13 (by default for SeqLoop-based sequences)
                noisefid(:,cChannel) = rawfid(:,idx,cCH);
                foundNoiseFid = true;
            elseif (isREFPHASESTABSCAN(idx))
                % skip the reference phase stabilization scans
            elseif (isPHASESTABSCAN(idx))
                % skip the phase stabilization scans
            else
                if (isSpec)
                    if (isPHASCOR(idx))
                        % save the phase correction lines separately (summed)
                        [reffid(:,cChannel), avgWarnStatus_spcpcor] = accu_data(reffid(:,cChannel), rawfid(:,idx,cCH), defaultAverage, avgWarnStatus_spcpcor, 'spec phascor');
                    else
                        fid(:,cChannel,cSet,cEcho,cTimePt,cImgLine,cPhase,cSeg) = rawfid(:,idx,cCH);
                    end
                else
                    if (isPHASCOR(idx))
                        % save the phase correction lines separately for epi
                        if (lastwaspcor(cSlice) == 0)
                            pcorctr(cSlice) = pcorctr(cSlice) + 1;
                        end
                        lastwaspcor(cSlice) = 1;
                        if (isEPI)
                            avgPCor = 1; % -1 = do not average 1st,3rd pcor lines; +1 = do average (Siemens default)
                            if (isSINGLEBANDREF(idx))
                                [sbpcor.fid(:,pcorctr(cSlice),cSeg,cSlice,cChannel), avgWarnStatus_sbpcor] = accu_data(sbpcor.fid(:,pcorctr(cSlice),cSeg,cSlice,cChannel), rawfid(:,idx,cCH), avgPCor, avgWarnStatus_sbpcor, 'sbref phascor'); % sum them
                                if ((cAcq == 2) && (avgPCor == 1)), sbpcor.fid(:,pcorctr(cSlice),cSeg,cSlice,cChannel) = sbpcor.fid(:,pcorctr(cSlice),cSeg,cSlice,cChannel) / 2; end % average the first line which is acquired twice
                            else
                                [pcor.fid(:,pcorctr(cSlice),cSeg,cSlice,cChannel), avgWarnStatus_mbpcor] = accu_data(pcor.fid(:,pcorctr(cSlice),cSeg,cSlice,cChannel), rawfid(:,idx,cCH), avgPCor, avgWarnStatus_mbpcor, 'mb phascor'); % sum them
                                if ((cAcq == 2) && (avgPCor == 1)), pcor.fid(:,pcorctr(cSlice),cSeg,cSlice,cChannel) = pcor.fid(:,pcorctr(cSlice),cSeg,cSlice,cChannel) / 2; end % average the first line which is acquired twice
                                %fprintf('idx %d = pcor: pcorctr %d; seg %d; slice %d; cha %d; cCH %d\n', idx, pcorctr(cSlice), cSeg, cSlice, cChannel, cCH);
                            end
                        end
                    else
                        % patref line
                        if ( isPATREFSCAN(idx) && ~isPATREFANDIMASCAN(idx) )
                            if (isEPI)
                                if (isSINGLEBANDREF(idx))
                                    sbpatref.pcor_idx(:,cPatLine,cSlice,cPart,cEcho) = [pcorctr(cSlice) cSeg];
                                else
                                    patref.pcor_idx(:,cPatLine,cSlice,cPart,cEcho) = [pcorctr(cSlice) cSeg];
                                end
                            end
                            if (isSINGLEBANDREF(idx))
                                [sbpatref.fid(:,cPatLine,cSlice,cPart,cEcho,cChannel), avgWarnStatus_sbpatref] = accu_data(sbpatref.fid(:,cPatLine,cSlice,cPart,cEcho,cChannel), rawfid(:,idx,cCH), defaultAverage, avgWarnStatus_sbpatref, 'sbref patref');
                            else
                                [patref.fid(:,cPatLine,cSlice,cPart,cEcho,cChannel), avgWarnStatus_mbpatref] = accu_data(patref.fid(:,cPatLine,cSlice,cPart,cEcho,cChannel), rawfid(:,idx,cCH), defaultAverage, avgWarnStatus_mbpatref, 'mb patref');
                            end
                        else
                            % normal image line or combined image/patref line
                            if (~isNewEPI), sSlicePos(cSlice,:) = [mdh.flSag(idx) mdh.flCor(idx) mdh.flTra(idx)]; end
                            if (isEPI)
                                if (isSINGLEBANDREF(idx))
                                    fprintf(''); % this magically keeps MATLAB from crashing with OUT OF MEMORY error for some reason
                                    sbpcor.idx(:,cImgLine,cSlice,cPart,cEcho) = [pcorctr(cSlice) cSeg];
                                else
                                    pcor.idx(:,cImgLine,cSlice,cPart,cEcho,cTimePt) = [pcorctr(cSlice) cSeg];
                                end
                            end
                            if (isSINGLEBANDREF(idx))
                                [sbfid(:,cImgLine,cSlice,cPart,cEcho,cChannel), avgWarnStatus_sbimg] = accu_data(sbfid(:,cImgLine,cSlice,cPart,cEcho,cChannel), rawfid(:,idx,cCH), defaultAverage, avgWarnStatus_sbimg, 'sbref image line');
                            else
                                [fid(:,cImgLine,cSlice,cPart,cEcho,cChannel,cTimePt), avgWarnStatus_mbimg] = accu_data(fid(:,cImgLine,cSlice,cPart,cEcho,cChannel,cTimePt), rawfid(:,idx,cCH), defaultAverage, avgWarnStatus_mbimg, 'mb image line');
                            end
                        end
                        
                        lastwaspcor = 0*lastwaspcor;
                    end
                end
            end
        end
    end
    
    % update waitbar periodically
    if (mod(idx,1024) == 0), fprintf('\b\b\b\b%3d%%',round(100*idx/nTraces)); end
end
fprintf('\b\b\b\b100%%\n');

if (~foundNoiseFid), noisefid = []; end

% % sort slices (match to mrprot slice positions by minimizing distance)
% % skip this for new epi seq which stores these in the TD field already
% if (~isSpec && ~isNewEPI && (nSlices > 1))
%     fprintf('Shuffling slices to match mrprot slice sort order\n');
%     sliceIdx = zeros(nSlices);
%     for x=1:nSlices % loop for mrprot slices
%         if isfield(mrprot.sSliceArray.asSlice(x), 'sPosition')
%             local_dSag = getmrprot(mrprot.sSliceArray.asSlice(x).sPosition,'dSag');
%             local_dCor = getmrprot(mrprot.sSliceArray.asSlice(x).sPosition,'dCor');
%             local_dTra = getmrprot(mrprot.sSliceArray.asSlice(x).sPosition,'dTra');
%         else
%             local_dSag = 0; local_dCor = 0; local_dTra = 0;
%         end
%         for y=1:nSlices % loop for mdh slices
%             sliceIdx(x,y) = (local_dSag - sSlicePos(y,1))^2 + (local_dCor - sSlicePos(y,2))^2 + (local_dTra - sSlicePos(y,3))^2;
%         end
%     end
%     sliceCrossRef = zeros(2,nSlices,'uint16');
%     sliceCrossRef(1,:) = uint16(getmrprot(mrprot.sSliceArray,'anPos') + 1); % mrprot slices
%     for x=1:nSlices
%         [~,sliceCrossRef(2,x)] = min(sliceIdx(:,x));
%     end
%     
%     fid(:,:,sliceCrossRef(2,:),:,:,:,:) = fid(:,:,sliceCrossRef(1,:),:,:,:,:);
% end

% fill the return struct
if (exist('coil_names','var'))
    refs.coil_names = coil_names;
end
refs.noisefid = noisefid;
if (foundpat)
    refs.patref = patref;
end
if (isEPI)
    % for EPI+multi-band, varargout = {coil_names, noisefid, pcor, sbfid, sbpcor[, patref, sbpatref]}
    refs.pcor = pcor;
    if (isSBRef)
        refs.sbfid = sbfid;
        refs.sbpcor = sbpcor;
        if (foundpat)
            refs.sbpatref = sbpatref;
        end
    end
elseif (isRadial)
    refs.angles = angles;
elseif (isSpec)
    refs.reffid = reffid;
end

% clean up and finish
fprintf('Done!\n\n');





% ------------------------------------------------------------------------------------------------------------------------
function [out_data, new_warned_status] = accu_data(old_data, in_data, accu_flag, warned_status, warn_msg)
% function to copy or sum incoming data
% accu_flag == -1  -> expecting repeat lines, don't sum
% accu_flag ==  0  -> not expecting repeat lines, but warn and sum if encountered
% accu_flag == +1  -> expecting repeat lines (averages), sum and don't complain

new_warned_status = warned_status;
if (accu_flag == -1)
    out_data = in_data;
elseif (accu_flag == 1)
    out_data = old_data + in_data;
else
    if ((~warned_status) && (nnz(old_data)))
        fprintf('\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\bWARNING: %s found unexpected extra averages! Something is wrong! Summing anyway...\n', warn_msg);
        fprintf('Sorting raw data:   0%%');
        new_warned_status = true;
    end
    out_data = old_data + in_data;
end


% ------------------------------------------------------------------------------------------------------------------------
function outval = getmrprot(struct_in, fieldnm)
% wrapper function to safely read numeric mrprot parameter
% (since mrprot field names will not exist if the value is zero)

outval = 0; % default

if isstruct(struct_in)
    if isfield(struct_in, fieldnm)
        outval = struct_in.(fieldnm);
    end
else
    %error('getmrprot error: input was not a struct!');
end
